"""Top-level package for GisMap."""

from importlib.metadata import metadata

from gismap.gisgraphs.widget import GismapWidget as GismapWidget
from gismap.gismo import make_gismo as make_gismo
from gismap.search import Search as Search
from gismap.search import SearchDocuments as SearchDocuments
from gismap.search import SearchFeatures as SearchFeatures
from gismap.search import SearchLandmarks as SearchLandmarks
from gismap.search import search_to_html as search_to_html
from gismap.search import search_to_text as search_to_text
from gismap.sources.dblp import DBLP as DBLP
from gismap.sources.dblp import DBLPAuthor as DBLPAuthor
from gismap.sources.hal import HAL as HAL
from gismap.sources.hal import HALAuthor as HALAuthor
from gismap.sources.ldb import LDB as LDB
from gismap.sources.ldb import LDBAuthor as LDBAuthor
from gismap.utils.common import get_classes as get_classes

infos = metadata(__name__)
__version__ = infos["Version"]
__author__ = """Fabien Mathieu"""
__email__ = "fabien.mathieu@normalesup.org"



import json
import uuid

from domonic import tags

from gismap.gisgraphs.graph import lab_to_graph
from gismap.gisgraphs.groups import auto_groups, make_legend
from gismap.gisgraphs.js import default_script
from gismap.gisgraphs.options import edges as def_edges
from gismap.gisgraphs.options import interaction as def_interaction
from gismap.gisgraphs.options import nodes as def_nodes
from gismap.gisgraphs.options import physics as def_physics
from gismap.gisgraphs.style import default_style

default_vis_url = '"https://cdn.jsdelivr.net/npm/vis-network/standalone/esm/vis-network.min.js"'

gislink = tags.a(
    "&copy; GisMap 2025",
    _href="https://balouf.github.io/gismap/",
    _target="_blank",
    _class="watermark gislink",
)


def make_vis(lab, **kwargs):
    """
    Generate HTML visualization of a lab's collaboration network.

    Parameters
    ----------
    lab: :class:`~gismap.lab.labmap.LabMap`
        Lab to display.
    uid: :class:`str`, optional
        Unique identifier for DOM elements. Auto-generated if not provided.
    vis_url: :class:`str`, optional
        URL to vis-network library.
    groups: :class:`dict`, optional
        Group styling configuration.
    draw_legend: :class:`bool`, optional
        Whether to draw the legend. Defaults to True if multiple groups.
    physics: :class:`dict`, optional
        Physics engine configuration.
    nodes_options: :class:`dict`, optional
        Node styling options.
    edges_options: :class:`dict`, optional
        Edge styling options.
    interaction_options: :class:`dict`, optional
        Interaction settings.
    style: :class:`string.Template`, optional
        CSS template.
    script: :class:`string.Template`, optional
        JavaScript template.

    Returns
    -------
    :class:`str`
        HTML code as a string.
    """
    uid = kwargs.pop("uid", None)
    if uid is None:
        uid = str(uuid.uuid4())[:8]
    vis_url = kwargs.pop("vis_url", default_vis_url)
    groups = kwargs.pop("groups", None)
    groups = auto_groups(lab, groups)
    draw_legend = kwargs.pop("draw_legend", len(groups) > 1)
    physics = kwargs.pop("physics", def_physics)
    nodes_options = kwargs.pop("nodes_options", def_nodes)
    edges_options = kwargs.pop("edges_options", def_edges)
    interaction_options = kwargs.pop("interaction_option", def_interaction)
    options = {
        "physics": physics,
        "groups": groups,
        "nodes": nodes_options,
        "edges": edges_options,
        "interaction": interaction_options,
    }
    style = kwargs.pop("style", default_style)
    script = kwargs.pop("script", default_script)
    if kwargs:
        raise TypeError(f"unexpected keyword arguments: {repr(kwargs)}")

    nodes, edges = lab_to_graph(lab)

    parameters = {
        "vis_url": vis_url,
        "uid": uid,
        "nodes": json.dumps(nodes),
        "edges": json.dumps(edges),
        "options": json.dumps(options),
    }
    div = tags.div(_class="gisgraph", _id=f"box-{uid}")
    div.appendChild(tags.div(_id=f"vis-{uid}"))
    div.appendChild(gislink)
    div.appendChild(tags.button("Redraw()", _id=f"redraw-{uid}", _class="watermark button redraw"))
    div.appendChild(tags.button("Full Screen", _id=f"fullscreen-{uid}", _class="watermark button fullscreen"))
    comets = not all(n.get("connected") for n in nodes)
    if draw_legend or comets:
        div.appendChild(make_legend(groups, uid))
    modal = tags.div(_class="modal", _id=f"modal-{uid}")
    modal_content = tags.div(_class="modal-content")
    modal_content.appendChild(tags.span("&times;", _class="close", _id=f"modal-close-{uid}"))
    modal_content.appendChild(tags.div(_id=f"modal-body-{uid}"))
    modal.appendChild(modal_content)
    div.appendChild(modal)

    style = tags.style(style.substitute(**parameters))
    script = tags.script(script.substitute(**parameters), _type="module")

    return "\n".join(f"{content}" for content in [div, style, script])



import unicodedata

from bof.fuzz import Process


class Corrector:
    """
    A simple word corrector base on input vocabulary. Short words are discarded.

    Parameters
    ----------
    voc: :class:`list`
        Words (each entry may contain multiple words).
    score_cutoff: :class:`int`, default=20
        Threshold for correction.
    min_length: :class:`int`, default=3
        Minimal number of characters for correction to kick in.

    Examples
    --------

    >>> vocabulary = ['My Taylor Swift is Rich']
    >>> phrase = "How riche ise Tailor Swyft"
    >>> cor = Corrector(vocabulary, min_length=4)
    >>> cor(phrase)
    'How rich ise taylor swift'
    >>> cor = Corrector(vocabulary, min_length=2)
    >>> cor(phrase)
    'How rich is taylor swift'
    """

    def __init__(self, voc, score_cutoff=20, min_length=3):
        self.voc = {k.lower() for w in voc for k in w.split() if len(k) >= min_length}
        self.cutoff = score_cutoff
        self.min_length = min_length
        self.p = Process()
        self.p.fit(list(self.voc))

    def correct_word(self, w):
        if len(w) < self.min_length or w.lower() in self.voc:
            return w
        ww = self.p.extractOne(w, score_cutoff=self.cutoff)
        return ww[0] if ww is not None else w

    def __call__(self, text):
        return " ".join(self.correct_word(w) for w in text.split())


def reduce_keywords(kws):
    """
    Remove redundant subparts.

    Parameters
    ----------
    kws: :class:`list`
        List of words / co-locations.

    Returns
    -------
    :class:`list`
        Reduced list

    Examples
    --------

    >>> reduce_keywords(['P2P', 'Millimeter Waves', 'Networks', 'P2P Networks', 'Waves'])
    ['Millimeter Waves', 'P2P Networks']
    """
    indices = []
    for i, kw1 in enumerate(kws):
        accept = True
        for j, kw2 in enumerate(kws):
            if j != i and kw1 in kw2:
                accept = False
                break
        if accept:
            indices.append(i)
    return [kws[i] for i in indices]


def clean_aliases(name, alias_list):
    """
    Parameters
    ----------
    name: :class:`str`
        Main name.
    alias_list: :class:`list` or :class:`set`
        Aliases.

    Returns
    -------
    :class:`list`
        Aliases deduped, sorted, and with main name removed.
    """
    return sorted(set(n for nn in alias_list for n in [nn, asciify(nn)] if n != name))


def asciify(text):
    """
    Parameters
    ----------
    text: :class:`str`
        Some text (typically names) with annoying accents.

    Returns
    -------
    :class:`str`
        Same text simplified into ascii.

    Examples
    --------
    >>> asciify('Ana Bušić')
    'Ana Busic'
    >>> asciify("Thomas Deiß")
    'Thomas Deiss'
    """
    text = text.replace("ß", "ss")
    decomposed = unicodedata.normalize("NFD", text)
    no_accents = "".join(c for c in decomposed if unicodedata.category(c) != "Mn")
    ascii_text = no_accents.encode("ascii", "ignore").decode()
    return ascii_text


def normalized_name(txt):
    """
    Try to normalize names for facilitating comparisons. Name is lowered, split, asciified, sorted, and filtered.

    Parameters
    ----------
    txt: :class:`str`

    Returns
    -------
    :class:`str`

    Examples
    --------

    >>> normalized_name("Thomas Deiß")
    'deiss thomas'
    >>> normalized_name("Dario Rossi 001")
    'dario rossi'
    >>> normalized_name("James W. Roberts")
    'james roberts'
    """
    return " ".join(
        sorted(
            asciify(a) for a in txt.lower().replace("-", " ").split() if not (a.isdigit() or (len(a) < 3 and "." in a))
        )
    )


def normalized_title(txt):
    """
    Try to normalize titles for facilitating comparisons. Title is lowercased, asciified, and stripped of punctuation.

    Parameters
    ----------
    txt: :class:`str`

    Returns
    -------
    :class:`str`

    Examples
    --------

    >>> normalized_title("An Efficient Algorithm for P2P Networks")
    'an efficient algorithm for p2p networks'
    >>> normalized_title("A Study on the Use of Millimeter Waves: 5G Networks.")
    'a study on the use of millimeter waves 5g networks'
    """
    return " ".join("".join(c if c.isalnum() else " " for c in asciify(txt.lower())).split())


def auto_int(txt):
    try:
        return int(txt)
    except ValueError:
        return txt



from string import Template

# language=css
default_style = Template("""

                         #box-$uid {
                             position: relative;
                             width: 100% !important;
                             height: 80vh !important;
                             max-width: 100vw !important;
                             max-height: 80vh !important;
                             min-height: 80vh;
                             color: #111;
                             background-color: var(--pst-color-background, var(--jp-layout-color0, white));
                         }

                         #box-$uid:fullscreen {
                             height: 100vh !important;
                             max-height: 100vh !important;
                             min-height: 100vh;
                         }

                         #vis-$uid {
                             height: 100%; /* Make the inner div fill the parent */
                             width: 100%; /* Make the inner div fill the parent */
                             box-sizing: border-box;
                             border: 1px solid #444;
                         }

                         html[data-theme="dark"] #vis-$uid {
                             background-color: var(--pst-color-background, #14181e);
                         }

                         .modal {
                             display: none;
                             position: fixed;
                             z-index: 1000;
                             left: 0;
                             top: 0;
                             width: 100%;
                             height: 100%;
                             overflow: auto;
                             background-color: rgba(10, 10, 10, 0.85);
                         }

                         .modal-content {
                             background-color: #f4f4f7;
                             color: #222235;
                             margin: 10% auto;
                             padding: 24px;
                             border: 1px solid #888;
                             width: 50%;
                             border-radius: 8px;
                             box-shadow: 0 5px 15px rgba(0, 0, 0, .6);
                         }
                         .modal a {color: #2958d7;}
                         .modal a:visited {color: #8435a8;}

                         .close {
                             color: #777;
                             float: right;
                             font-size: 28px;
                             font-weight: bold;
                             cursor: pointer;
                         }

                         .close:hover, .close:focus {
                             color: #aaa;
                             text-decoration: none;
                             cursor: pointer;
                         }

                         .watermark {
                             position: absolute;
                             text-decoration: none;
                             color: #888;
                             font-size: min(2vw, 10px);
                             z-index: 10;
                         }

                         .gislink {
                             left: 10px;
                             bottom: 10px;
                             pointer-events: auto;
                         }

                         .button {
                             background: none;
                             border: none;
                             padding: 0;
                             margin: 0;
                             cursor: pointer;
                         }

                         .redraw {
                             left: 10px;
                             top: 10px;
                         }

                         .fullscreen {
                             bottom: 10px;
                             right: 10px;
                         }

                         .legend {
                             display: inline-block;
                             padding: 10px 16px;
                             border-radius: 8px;
                             box-shadow: 0 2px 8px 0 rgba(0, 0, 0, 0.10);
                             border: 1px solid var(--legend-border, #bbb);
                             background: var(--jp-layout-color1, #f5f5fa);
                             background-color: var(--legend-bg, rgba(240, 240, 245, 0.95));
                             position: absolute;
                             top: 12px;
                             right: 12px;
                             z-index: 20;
                         }

                         .legend-entry, .comet-entry {
                             display: flex;
                             margin-right: 10px;
                             align-items: center;
                             cursor: pointer;
                         }
                         """)



import re

# editorials = re.compile(r"ditorial|foreword", re.IGNORECASE)
# charlatans = re.compile(r"Raoult|Kofman|Buob")

editorials = ["ditorial", "Foreword", "Brief Announcement", "Preface", "Préface"]
charlatans = ["Raoult", "Kofman", "Buob"]


def re_filter(words):
    """
    Parameters
    ----------
    words: :class:`list` or :class:`str`
        List of word(s) to filter.

    Returns
    -------
    callable
        Filter function.
    """
    if isinstance(words, str):
        taboo = re.compile(words)
    else:
        taboo = re.compile("|".join(words))
    return lambda txt: taboo.search(txt) is None


def publication_size_filter(n_max=9):
    """
    Parameters
    ----------
    n_max: int, default=9
        Maximum number of co-authors allowed.

    Returns
    -------
    callable
        Filter on number of co-authors.
    """
    return lambda p: len(p.authors) <= n_max


def publication_oneword_filter(n_min=2):
    """

    Parameters
    ----------
    n_min: int, default=2
        Minimum number of words required in the title.

    Returns
    -------
    callable
        Filter on number of words required in the title.
    """
    return lambda p: len(p.title.split()) >= n_min


def publication_taboo_filter(w=None):
    """
    Parameters
    ----------
    w: :class:`list`, optional
        List of words to filter.

    Returns
    -------
    Callable
        Filter function on publications.
    """
    if w is None:
        w = editorials
    regex = re_filter(w)
    return lambda p: regex(p.title)


def author_taboo_filter(w=None):
    """
    Parameters
    ----------
    w: :class:`list`, optional
        List of words to filter.

    Returns
    -------
    Callable
        Filter function on authors.
    """
    if w is None:
        w = charlatans
    regex = re_filter(w)
    return lambda p: regex(p.name)



import re
import zlib
from contextlib import contextmanager
from pathlib import Path

from tqdm.auto import tqdm

from gismap.sources.dblp import DBLP_TYPES
from gismap.utils.requests import session

key_re = r"<https://dblp.org/rec/([^>]+)>"
title_re = r'.*?dblp:title\s+"([^"]+)"'
type_re = r".*?dblp:bibtexType\s+bibtex:(\w+)"
authors_re = r".*?dblp:hasSignature\s+(\[.*\])\s*;"
url_re = r"(?:.*?dblp:primaryDocumentPage <([^>]+)>)?"
stream_re = r"(?:.*?dblp:publishedInStream ([^;]+) ;)?"
pages_re = r'(?:.*?dblp:pagination "([^"]+)")?'
venue_re = r'(?:.*?dblp:publishedIn\s+"([^"]+?)")?'
year_re = r'.*?"(\d{4})"\^\^<http://www.w3.org/2001/XMLSchema#gYear>'

pub_re = re.compile(
    "".join(
        [
            key_re,
            title_re,
            type_re,
            authors_re,
            url_re,
            stream_re,
            pages_re,
            venue_re,
            year_re,
        ]
    ),
    flags=re.S,
)

streams_re = re.compile(r"<https://dblp.org/streams/((?:conf|journals)/[^>]+)>")

authid_re = re.compile(
    r'\[.*?signatureDblpName\s*?"([^"]+?)(?:\s+\d+)?".*?signatureCreator\s*<https://dblp.org/pid/([^>]+?)>.*?]',
    flags=re.S,
)


def parse_block(dblp_block):
    """
    Parameters
    ----------
    dblp_block: :class:`str`
        A DBLP publication, turtle format.

    Returns
    -------
    key: :class:`str`
        DBLP key.
    title: :class:`str`
        Publication title.
    type: :class:`str`
        Type of publication.
    authors: :class:`dict`
        Publication authors (key -> name)
    url: :class:`str` or :class:`NoneType`
        Publication URL.
    stream: :class:`list` or :class:`NoneType`
        Publication streams (normalized journal/conf).
    pages: :class:`str` or :class:`NoneType`
        Publication pages.
    venue: :class:`str` or :class:`NoneType`
        Publication venue (conf/journal).
    year: :class:`int`
        Year of publication.
    """
    items = pub_re.search(dblp_block)
    if items is None:
        return None
    key, title, typ, authors, url, stream, pages, venue, year = items.groups()
    typ = typ.lower()
    typ = DBLP_TYPES.get(typ, typ)
    if stream:
        stream = streams_re.findall(stream)
    authors = {i: n for n, i in authid_re.findall(authors)}
    if authors:
        return key, title, typ, authors, url, stream, pages, venue, int(year)
    return None


@contextmanager
def get_stream(source, chunk_size=1024 * 64):
    """
    Parameters
    ----------
    source: :class:`str` or :class:`~pathlib.Path`
        Where the content. Can be on a local file or on the Internet.
    chunk_size: :class:`int`, optional
        Desired chunk size. For streaming gz content, must be a multiple of 32kB.

    Yields
    -------
    iterable
        Chunk iterator that streams the content.
    :class:`int`
        Source size (used later to compute ETA).
    """
    if isinstance(source, str) and source.startswith("https://"):
        # URL HTTP
        with session.get(source, stream=True) as r:
            r.raise_for_status()
            total = int(r.headers.get("content-length", 0)) or None
            yield r.iter_content(chunk_size=chunk_size), total
    else:
        source = Path(source)
        if not source.exists():
            yield [], 0
            return None
        total = source.stat().st_size
        with source.open("rb") as file_handle:

            def read_chunks():
                while True:
                    chunk = file_handle.read(chunk_size)
                    if not chunk:
                        break
                    yield chunk

            yield read_chunks(), total


def publis_streamer(source, chunk_size=1024 * 64, encoding="unicode_escape"):
    """
    Parameters
    ----------
    source: :class:`str` or :class:`~pathlib.Path`
        Where the DBLP turtle content is. Can be on a local file or on the Internet.
    chunk_size: :class:`int`, optional
        Desired chunk size. Must be a multiple of 32kB.
    encoding: :class:`str`, default=unicode_escape
        Encoding of stream.

    Yields
    -------
    key: :class:`str`
        DBLP key.
    title: :class:`str`
        Publication title.
    type: :class:`str`
        Type of publication.
    authors: :class:`dict`
        Publication authors (key -> name).
    venue: :class:`str`
        Publication venue (conf/journal).
    year: :class:`int`
        Year of publication.
    """
    with get_stream(source, chunk_size=chunk_size) as (stream, total):
        with tqdm(total=total, unit="B", unit_scale=True, unit_divisor=1024, desc="Processing") as pbar:
            decomp = zlib.decompressobj(16 + zlib.MAX_WBITS)
            text_buffer = ""
            for chunk in stream:
                if not chunk:
                    continue

                pbar.update(len(chunk))
                data = decomp.decompress(chunk)
                if not data:
                    continue
                text_buffer += data.decode(encoding, errors="replace")

                blocks = text_buffer.split("\n\n")
                text_buffer = blocks[-1]
                for block in blocks[:-1]:
                    pub = parse_block(block)
                    if pub:
                        yield pub

        data = decomp.flush()
        if data:
            text_buffer += data.decode(encoding, errors="replace")

        if text_buffer:
            blocks = text_buffer.split("\n\n")
            for block in blocks:
                pub = parse_block(block)
                if pub:
                    yield pub



from dataclasses import dataclass, field
from time import sleep
from typing import ClassVar
from urllib.parse import quote_plus

from bs4 import BeautifulSoup as Soup

from gismap.sources.models import DB, Author, Publication
from gismap.utils.requests import get
from gismap.utils.text import auto_int, clean_aliases


@dataclass(repr=False)
class DBLP(DB):
    db_name: ClassVar[str] = "dblp"
    author_backoff: ClassVar[float] = 5.0
    publi_backoff: ClassVar[float] = 5.0

    @classmethod
    def search_author(cls, name, wait=True):
        """
        Parameters
        ----------
        name: :class:`str`
            People to find.
        wait: :class:`bool`, default=True
            Wait a bit to avoid 429.

        Returns
        -------
        :class:`list`
            Potential matches.

        Examples
        --------

        >>> fabien = DBLP.search_author("Fabien Mathieu")
        >>> fabien  # doctest: +FLAKY
        [DBLPAuthor(name='Fabien Mathieu', key='66/2077')]
        >>> fabien[0].url  # doctest: +FLAKY
        'https://dblp.org/pid/66/2077.html'
        >>> manu = DBLP.search_author("Manuel Barragan")
        >>> manu  # doctest: +FLAKY, +NORMALIZE_WHITESPACE
        [DBLPAuthor(name='Manuel Barragan', key='07/10587'),
        DBLPAuthor(name='Manuel Barragan', key='83/3865'),
        DBLPAuthor(name='Manuel Barragan', key='188/0198')]
        >>> DBLP.search_author("NotaSearcherName", wait=False)  # doctest: +FLAKY
        []
        """
        dblp_api = "https://dblp.org/search/author/api"
        dblp_args = {"q": name}
        r = get(dblp_api, params=dblp_args)
        soup = Soup(r, features="xml")
        if wait:
            sleep(cls.author_backoff)
        return [
            DBLPAuthor(
                name=name,
                key=hit.url.text.split("pid/")[1],
                aliases=clean_aliases(name, [hit.author.text] + [alia.text for alia in hit("alias")]),
            )
            for hit in soup("hit")
        ]

    @classmethod
    def from_author(cls, a, wait=True):
        """
        Returns
        -------
        :class:`list`
            Papers available in DBLP.
        wait: :class:`bool`
            Wait a bit to avoid 429.
        """
        r = get(f"https://dblp.org/pid/{a.key}.xml")
        soup = Soup(r, features="xml")
        if wait:
            sleep(cls.publi_backoff)
        res = [DBLPPublication.from_soup(r) for r in soup("r")]
        return [p for p in res if p.authors]


@dataclass(repr=False)
class DBLPAuthor(Author, DBLP):
    """
    Examples
    --------

    >>> fabien = DBLPAuthor('Fabien Mathieu', key='66/2077')
    >>> publications = sorted(fabien.get_publications(),
    ...                 key=lambda p: p.title)
    >>> publications[0].url  # doctest: +FLAKY, +NORMALIZE_WHITESPACE
     'https://dblp.org/rec/conf/iptps/BoufkhadMMPV08.html'
    >>> publications[-1]  # doctest: +FLAKY, +NORMALIZE_WHITESPACE
    DBLPPublication(title='Upper Bounds for Stabilization in Acyclic Preference-Based Systems.',
    authors=[DBLPAuthor(name='Fabien Mathieu', key='66/2077')], venue='SSS', type='conference', year=2007,
    key='conf/sss/Mathieu07')

    """

    key: str
    aliases: list = field(default_factory=list)

    @property
    def url(self):
        if self.key:
            return f"https://dblp.org/pid/{self.key}.html"
        return f"https://dblp.org/search?q={quote_plus(self.name)}"

    def get_publications(self, wait=True):
        return DBLP.from_author(self, wait=wait)


DBLP_TYPES = {
    "article": "journal",
    "inproceedings": "conference",
    "proceedings": "book",
    "informal": "report",
    "phdthesis": "thesis",
    "habil": "hdr",
    "software": "software",
}


@dataclass(repr=False)
class DBLPPublication(Publication, DBLP):
    """
    Publication from the DBLP database.

    Parameters
    ----------
    title: :class:`str`
        Publication title.
    authors: :class:`list`
        List of :class:`DBLPAuthor` objects.
    venue: :class:`str`
        Publication venue.
    type: :class:`str`
        Publication type.
    year: :class:`int`
        Publication year.
    key: :class:`str`
        DBLP record key.
    metadata: :class:`dict`
        Additional metadata (pages, volume, etc.).
    """

    key: str
    metadata: dict = field(default_factory=dict)

    @property
    def url(self):
        if self.key:
            return f"https://dblp.org/rec/{self.key}.html"
        else:
            return None

    @classmethod
    def from_soup(cls, soup):
        p = soup.find()
        typ = p.get("publtype", p.name)
        typ = DBLP_TYPES.get(typ, typ)

        res = {
            "type": typ,
            "key": p["key"],
            "title": p.title.text,
            "year": int(p.year.text),
        }
        for tag in ["booktitle", "journal"]:
            t = p.find(tag)
            if t:
                res["venue"] = t.text
                break
        else:
            res["venue"] = "unpublished"
        res["authors"] = [DBLPAuthor(key=a["pid"], name=a.text) for a in p("author")]

        metadata = dict()
        for tag in p.find_all(recursive=False):
            name = tag.name
            if name not in {"title", "year", "author", "booktitle", "journal"}:
                metadata[name] = auto_int(tag.text)

        return cls(**res, metadata=metadata)



import doctest
import warnings

import pytest

FLAKY = doctest.register_optionflag("FLAKY")


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    report = outcome.get_result()
    if report.when == "call" and report.failed and hasattr(item, "dtest"):
        # Check if the failing example was marked # doctest: +FLAKY
        exc = call.excinfo.value
        example = getattr(exc, "example", None)
        if example is not None and example.options.get(FLAKY):
            warnings.warn(
                f"Flaky doctest failed: {item.nodeid}\n{report.longreprtext}",
                stacklevel=1,
            )
            report.outcome = "skipped"
            report.wasxfail = "Flaky doctest (network-dependent)"



from collections import Counter, defaultdict
from dataclasses import dataclass

import numpy as np
from bof.fuzz import Process

from gismap.lab.lab_author import LabAuthor
from gismap.sources.multi import sort_author_sources
from gismap.utils.text import normalized_name


@dataclass
class ProspectStrength:
    """
    Measures the interaction between an external author and a lab by counting co-authors and publications.

    A (max,+) addition is handled to deal with multiple keys.

    Examples
    --------

    >>> a1 = ProspectStrength(3, 5)
    >>> a2 = ProspectStrength(2, 10)
    >>> a1 > a2
    True
    >>> a1 + a2
    ProspectStrength(coauthors=3, publications=15)
    """

    coauthors: int
    publications: int

    def __call__(self):
        return self.coauthors, self.publications

    def __add__(self, other):
        if other == 0:
            return self
        return ProspectStrength(
            coauthors=max(self.coauthors, other.coauthors),
            publications=self.publications + other.publications,
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __lt__(self, other):
        return self() < other()


def count_prospect_entries(lab):
    """
    Associate to external coauthors (prospects) their lab strength.

    Parameters
    ----------
    lab: :class:`~gismap.lab.labmap.LabMap`
        Reference lab.

    Returns
    -------
    :class:`dict` of :class:`str` to :class:`~gismap.lab.expansion.ProspectStrength`
        Lab strengths.
    """
    count_coauthors = defaultdict(set)
    count_publications = []
    for p in lab.publications.values():
        for s in p.sources:
            new_authors = set()
            lab_authors = set()
            for a in s.authors:
                if hasattr(a, "db_name"):
                    new_authors.add(a.key)
                    count_publications.append(a.key)
                else:
                    lab_authors.add(a.key)
            for a in lab_authors:
                count_coauthors[a].update(new_authors)

    count_coauthors = Counter(k for new_authors in count_coauthors.values() for k in new_authors)
    count_publications = Counter(count_publications)

    return {
        k: ProspectStrength(coauthors=count_coauthors.get(k, 0), publications=count_publications[k])
        for k in count_publications
    }


class Prospect:
    """
    Candidate for integration to lab.

    Parameters
    ----------
    author: :class:`~gismap.sources.models.Author`
        Reference author. Must have a key.
    strengths: :class:`dict`
        Dictionary of ProspectStrength.
    """

    def __init__(self, author, strengths):
        self.name = normalized_name(author.name)
        self.author = author
        self.score = strengths[author.key]

    def __lt__(self, other):
        return self.score < other.score

    def __repr__(self):
        return f"Prospect({self.name}, key={self.author.key}, s={self.score()}"


def get_prospects(lab):
    """
    Parameters
    ----------
    lab: :class:`~gismap.lab.labmap.LabMap`
        Reference lab.

    Returns
    -------
    :class:`list` of :class:`~gismap.lab.expansion.Prospect`
        List of prospects.
    """
    strengths = count_prospect_entries(lab)
    prospect_dico = {
        a.key: a
        for p in lab.publications.values()
        for s in p.sources
        for a in s.authors
        if hasattr(a, "db_name") and all(f(a) for f in lab.author_selectors)
    }
    return [Prospect(a, strengths) for a in prospect_dico.values()]


@dataclass
class Member:
    """
    Basic information about a lab member for name matching.

    Parameters
    ----------
    name : :class:`str`
        Normalized name.
    key : :class:`str`
        Author key.
    """

    name: str
    key: str


def get_member_names(lab):
    """
    Parameters
    ----------
    lab: :class:`~gismap.lab.labmap.LabMap`
        Reference lab.

    Returns
    -------
    :class:`list`
        Tuples simplified-name -> key
    """
    return [(name, k) for k, a in lab.authors.items() for name in {normalized_name(n) for n in [a.name, *a.aliases]}]


def trim_sources(author):
    """
    Inplace reduction of sources, keeping one unique source per db.

    Parameters
    ----------
    author: :class:`~gismap.sources.multi.SourcedAuthor`
        An author.

    Returns
    -------
    None
    """
    sources = []
    seen = set()
    for s in author.sources:
        if s.db_name not in seen:
            seen.add(s.db_name)
            sources.append(s)
    author.sources = sources


def proper_prospects(lab, length_impact=0.05, threshold=80, n_range=4, max_new=None, trim=True):
    """
    Find and rank external collaborators for potential lab expansion.

    Identifies authors from publications who are not already lab members,
    groups them by name similarity, and ranks by collaboration strength.

    Parameters
    ----------
    lab : :class:`~gismap.lab.labmap.LabMap`
        Reference lab.
    length_impact : :class:`float`, default=0.05
        Length impact for name similarity matching.
    threshold : :class:`int`, default=80
        Similarity threshold for grouping authors.
    n_range : :class:`int`, default=4
        N-gram range for name comparison.
    max_new : :class:`int`, optional
        Maximum number of new authors to return.
    trim : :class:`bool`, default=True
        If True, keep only one source per database for each author.

    Returns
    -------
    :class:`tuple`
        (existing, new_rosetta) where existing maps external keys to lab member keys,
        and new_rosetta maps source keys to new LabAuthor objects.
    """
    member_names = get_member_names(lab)
    prospects = get_prospects(lab)

    if len(member_names) == 0 or len(prospects) == 0:
        return dict(), dict()

    done = np.zeros(len(prospects), dtype=bool)

    # Compare current and prospects names to re-attach ghosts
    p = Process(length_impact=length_impact, n_range=n_range)
    p.allow_updates = False
    p.fit([n[0] for n in member_names])
    jc = p.transform([prospect.name for prospect in prospects])
    best_choice = np.argmax(jc, axis=1)
    existing = dict()
    for i, j in enumerate(best_choice):
        if jc[i, j] > threshold:
            existing[prospects[i].author.key] = member_names[j][1]
            done[i] = True

    # Regroup remaining prospects
    p.reset()
    names = [prospect.name for prospect in prospects]
    p.fit(names)
    jc = p.transform(names)
    new_lab = []
    for i, prospect in enumerate(prospects):
        if done[i]:
            continue
        locs = [j for j in np.where(jc[i, :] > threshold)[0] if not done[j]]
        done[locs] = True
        sources = sort_author_sources([prospects[j].author for j in locs])
        if sources:
            strength = sum(prospects[j].score for j in locs)
            new_author = LabAuthor.from_sources(sources)
            new_lab.append((strength, new_author))

    # Extract top prospects
    new_lab = [a[1] for a in sorted(new_lab, key=lambda a: a[0], reverse=True)][:max_new]
    new_rosetta = {s.key: a for a in new_lab for s in a.sources}

    # Remove extra sources
    if trim:
        for a in new_lab:
            trim_sources(a)

    return existing, new_rosetta



from pathlib import Path

from gismo import MixInIO
from IPython.display import HTML, display
from tqdm.auto import tqdm

from gismap.gisgraphs.builder import make_vis
from gismap.lab.expansion import proper_prospects
from gismap.lab.filters import (
    author_taboo_filter,
    publication_oneword_filter,
    publication_size_filter,
    publication_taboo_filter,
)
from gismap.lab.lab_author import (
    AuthorMetadata,
    LabAuthor,
    db_dict,
    default_dbs,
    labify_publications,
)
from gismap.sources.multi import (
    regroup_authors,
    regroup_publications,
)
from gismap.utils.common import list_of_objects
from gismap.utils.logger import logger


class LabMap(MixInIO):
    """
    Abstract class for labs.

    Actual Lab classes can be created by implementing the `_author_iterator` method.

    Labs can be saved with the `dump` method and loaded with the `load` method.

    Parameters
    ----------
    name: :class:`str`
        Name of the lab. Can be set as class or instance attribute.
    dbs: :class:`list`, default=[:class:`~gismap.sources.hal.HAL`, :class:`~gismap.sources.ldb.LDB`]
        List of DB sources to use.


    Attributes
    -----------

    author_selectors: :class:`list`
        Author filters. Default: minimal filtering.
    publication_selectors: :class:`list`
        Publication filter. Default: less than 10 authors, not an editorial, at least two words in the title.
    """

    name = None
    dbs = default_dbs

    def __init__(self, name=None, dbs=None):
        if name is not None:
            self.name = name
        self.dbs = dbs
        self.author_selectors = [author_taboo_filter()]
        self.publication_selectors = [
            publication_size_filter(),
            publication_taboo_filter(),
            publication_oneword_filter(),
        ]
        self.authors = None
        self.publications = None

    def __repr__(self):
        return f"Lab {self.name}"

    def _author_iterator(self):
        """
        Yields
        ------
        :class:`~gismap.lab.lab_author.LabAuthor`
        """
        raise NotImplementedError

    def update_authors(self, desc="Author information"):
        """
        Populate the authors attribute (:class:`dict` [:class:`str`, :class:`~gismap.lab.lab_author.LabAuthor`]).

        Returns
        -------
        None
        """
        self.authors = dict()
        for author in tqdm(self._author_iterator(), desc=desc):
            if not all(f(author) for f in self.author_selectors):
                continue
            if len(author.sources) == 0:
                author.auto_sources(dbs=list_of_objects(self.dbs, db_dict(), default=default_dbs))
            if author.sources:
                self.authors[author.key] = author
            if author.metadata.img is None:
                author.auto_img()
            if author.metadata.group is None:
                author.metadata.group = self.name

    def update_publis(self, desc="Publications information"):
        """
        Populate the publications attribute
        (:class:`dict` [:class:`str`, :class:`~gismap.sources.multi.SourcedPublication`]).

        Returns
        -------
        None
        """
        pubs = dict()
        for author in tqdm(self.authors.values(), desc=desc):
            pubs.update(author.get_publications(clean=False, selector=self.publication_selectors))
        regroup_authors(self.authors, pubs)
        self.publications = regroup_publications(pubs)

    def expand(self, target=None, group="moon", desc="Moon information", **kwargs):
        if target is None:
            target = len(self.authors) // 3
        old, rosetta = proper_prospects(self, max_new=target, **kwargs)
        new = {a.key: a for a in rosetta.values()}
        for k, v in old.items():
            rosetta[k] = self.authors[v]
        logger.debug(f"{len(new)} new authors selected")
        if len(new) == 0:
            logger.warning("Expansion failed: no new author found.")
            return None

        self.authors.update(new)

        pubs = dict()
        for author in tqdm(new.values(), desc=desc):
            author.auto_img()
            author.metadata.group = group
            pubs.update(author.get_publications(clean=False, selector=self.publication_selectors))

        for pub in self.publications.values():
            for source in pub.sources:
                pubs[source.key] = source

        labify_publications(pubs.values(), rosetta)

        self.publications = regroup_publications(pubs)

        return None

    def html(self, **kwargs):
        """
        Generate HTML representation of the collaboration graph.

        Parameters
        ----------
        **kwargs
            Passed to :func:`~gismap.gisgraphs.builder.make_vis`.

        Returns
        -------
        :class:`str`
            HTML content as a string.
        """
        return make_vis(self, **kwargs)

    def save_html(self, name=None, **kwargs):
        """
        Save the collaboration graph as an HTML file.

        Parameters
        ----------
        name: :class:`str`, optional
            Output filename. Defaults to lab name.
        **kwargs
            Passed to :meth:`html`.

        Returns
        -------
        None
        """
        if name is None:
            name = self.name
        name = Path(name).with_suffix(".html")
        with open(name, "w", encoding="utf8") as f:
            f.write(self.html(**kwargs))

    def show_html(self, **kwargs):
        """
        Display the collaboration graph in a Jupyter notebook.

        Parameters
        ----------
        **kwargs
            Passed to :meth:`html`.

        Returns
        -------
        None
        """
        display(HTML(self.html(**kwargs)))


class ListMap(LabMap):
    """
    Simplest way to create a lab: with a list of names.

    Parameters
    ----------
    author_list: :class:`list` of :class:`str`
        List of authors names.
    args: :class:`list`
        Arguments to pass to the :class:`~gismap.lab.labmap.LabMap` constructor.
    kwargs: :class:`dict`
        Keyword arguments to pass to the :class:`~gismap.lab.labmap.LabMap` constructor.
    """

    def __init__(self, author_list, *args, **kwargs):
        self.author_list = author_list
        super().__init__(*args, **kwargs)

    def _author_iterator(self):
        for name in self.author_list:
            yield LabAuthor(name=name, metadata=AuthorMetadata())



from bs4 import BeautifulSoup as Soup

from gismap.lab import LabAuthor
from gismap.lab.lab_author import AuthorMetadata
from gismap.lab.labmap import LabMap
from gismap.utils.requests import get


def lamsade_parse(div):
    """
    Parameters
    ----------
    div: :class:`~bs4.BeautifulSoup`
        Soup of the div of one researcher

    Returns
    -------
    :class:`tuple`
        name, image url (or None), webpage (or None)
    """
    img = div.img["src"] if div.img and "default" not in div.img["src"] else None
    url = div.a["href"] if div.a else None
    name = div.h3.text.strip().title()
    # name = " ".join(name.split(" ", 1)[::-1])
    return name, img, url


class Lamsade(LabMap):
    """
    Class for handling the Lamsade team (Dauphine).
    The data is retrieved from the webpage of the team, and parsed with BeautifulSoup.
    The name, image url and webpage of each researcher are extracted and stored in a LabAuthor object.
    """

    name = "Lamsade"
    base_url = "https://www.lamsade.dauphine.fr/"
    directory = "fr/personnes/enseignants-chercheurs-et-chercheurs.html"

    def _author_iterator(self):
        soup = Soup(get(self.base_url + self.directory), features="lxml")
        for a in soup("div", class_="csc-frame"):
            name, img, url = lamsade_parse(a)
            img = self.base_url + img if img else None
            url = url if url else None
            yield LabAuthor(name=name, metadata=AuthorMetadata(url=url, img=img, group=self.name))



import re

from bs4 import BeautifulSoup as Soup

from gismap.lab.lab_author import AuthorMetadata, LabAuthor
from gismap.lab.labmap import LabMap
from gismap.utils.requests import get


class Lip6Map(LabMap):
    """
    Class for handling a LIP6 team using
    `https://www.lip6.fr/recherche/team_membres.php?acronyme=*team_acronym*` as entry point.
    Default to `NPA` team.
    """

    name = "NPA"

    def _author_iterator(self):
        url = f"https://www.lip6.fr/recherche/team_membres.php?acronyme={self.name}"
        soup = Soup(get(url), "lxml")
        for a in soup.table("a"):
            name = a.text.replace("\xa0", " ").strip()
            if not name:
                continue
            metadata = AuthorMetadata(group=self.name)
            previous = a.find_previous_sibling()
            if previous is not None and "user" in previous.get("class", []):
                metadata.url = previous["href"].strip()
            fiche = "https://www.lip6.fr/" + a["href"].split("/", 1)[1]
            img = Soup(get(fiche), "lxml").img
            if img and "reflet" in img["class"] and "noPhoto" not in img["src"]:
                metadata.img = "https://www.lip6.fr/" + img["src"].split("/", 1)[1]
            yield LabAuthor(name=name, metadata=metadata)


class Lip6Full(Lip6Map):
    """
    Class for handling all LIP6 teams using `https://www.lip6.fr/informations/annuaire.php` to get team names.
    """

    name = "LIP6"

    def _author_iterator(self):
        groups = re.compile(r'acronyme=(.*?)[\'"]')
        for group in groups.findall(get("https://www.lip6.fr/informations/annuaire.php")):
            yield from Lip6Map(name=group)._author_iterator()



from bs4 import BeautifulSoup as Soup

from gismap.lab.lab_author import AuthorMetadata, LabAuthor
from gismap.lab.labmap import LabMap
from gismap.utils.requests import get


def get_irif_teams():
    soup = Soup(get("https://www.irif.fr/"), "lxml")
    teams = [a for a in soup("ul", {"class": "nav"})[1]("a") if hasattr(a, "href") and "equipes" in a["href"]]
    return {a["href"].split("/")[-2]: a.text.strip() for a in teams}


irif_teams = get_irif_teams()


class IrifMap(LabMap):
    """
    Class for handling an IRIF team using `https://www.irif.fr/equipes/*team_acronym*/index` as entry point.
    Default to `graphes` team.
    """

    name = "graphes"

    def _author_iterator(self):
        url = f"https://www.irif.fr/equipes/{self.name}/index"
        soup = Soup(get(url), "lxml")
        for a in [td.a for td in soup("td", {"class": "col0"})]:
            if a is None:
                continue
            name = a.text.replace("\xa0", " ").strip()
            if not name:
                continue
            metadata = AuthorMetadata(group=irif_teams[self.name], url=a["href"])
            yield LabAuthor(name=name, metadata=metadata)


class IrifFull(IrifMap):
    """
    Class for handling all IRIF teams.
    """

    name = "IRIF"

    def _author_iterator(self):
        author_dict = dict()
        for team in irif_teams:
            for author in IrifMap(name=team)._author_iterator():
                if author.name in author_dict:
                    author_dict[author.name].metadata.group = "Polyteam"
                else:
                    author_dict[author.name] = author
        for author in author_dict.values():
            yield author



import json
from collections import defaultdict
from dataclasses import dataclass, field
from time import sleep
from typing import ClassVar
from urllib.parse import quote_plus

from bs4 import BeautifulSoup as Soup

from gismap.sources.models import DB, Author, Publication  #  DBAuthor, DBPublication
from gismap.utils.common import unlist
from gismap.utils.requests import get
from gismap.utils.text import clean_aliases


@dataclass(repr=False)
class HAL(DB):
    db_name: ClassVar[str] = "hal"
    author_backoff: ClassVar[float] = 0.5
    publi_backoff: ClassVar[float] = 0.5

    @classmethod
    def search_author(cls, name, wait=True):
        """
        Parameters
        ----------
        name: :class:`str`
            People to find.
        wait: :class:`bool`, default=True
            Wait a bit to avoid 429.

        Returns
        -------
        :class:`list`
            Potential matches.

        Examples
        --------

        >>> fabien = HAL.search_author("Fabien Mathieu")
        >>> fabien
        [HALAuthor(name='Fabien Mathieu', key='fabien-mathieu')]
        >>> fabien = fabien[0]
        >>> fabien.url
        'https://hal.science/search/index/?q=*&authIdHal_s=fabien-mathieu'
        >>> HAL.search_author("Laurent Viennot")[0]
        HALAuthor(name='Laurent Viennot', key='laurentviennot')
        >>> HAL.search_author("NotaSearcherName")
        []
        >>> HAL.search_author("Ana Busic")
        [HALAuthor(name='Ana Busic', key='anabusic')]
        >>> HAL.search_author("Potop-Butucaru Maria")  # doctest:  +NORMALIZE_WHITESPACE
        [HALAuthor(name='Potop-Butucaru Maria', key='841868', key_type='pid')]
        >>> diego = HAL.search_author("Diego Perino")
        >>> diego  # doctest:  +NORMALIZE_WHITESPACE
        [HALAuthor(name='Diego Perino', key='847558', key_type='pid'),
        HALAuthor(name='Diego Perino', key='978810', key_type='pid')]
        >>> diego[1].url
        'https://hal.science/search/index/?q=*&authIdPerson_i=978810'
        """
        hal_api = "https://api.archives-ouvertes.fr/ref/author/"
        fields = ",".join(["label_s", "idHal_s", "person_i", "fullName_s"])
        hal_args = {"q": name, "fl": fields, "wt": "json"}
        r = get(hal_api, params=hal_args)
        response = json.loads(r)["response"]
        hids = defaultdict(set)
        pids = defaultdict(set)
        names = set()
        for a in response.get("docs", []):
            if "label_s" in a:
                if "idHal_s" in a:
                    hids[a["idHal_s"]].add(a.get("label_s"))
                elif "person_i" in a:
                    pids[a["person_i"]].add(a.get("label_s"))
            elif "fullName_s" in a:
                names.add(a["fullName_s"])
        res = [HALAuthor(name=name, key=k, aliases=clean_aliases(name, v)) for k, v in hids.items()] + [
            HALAuthor(name=name, key=str(k), aliases=clean_aliases(name, v), key_type="pid") for k, v in pids.items()
        ]
        if wait:
            sleep(cls.author_backoff)
        return (
            res
            if res
            else [
                HALAuthor(
                    name=name,
                    key=fullname,
                    aliases=[],
                    key_type="fullname",
                )
                for fullname in names
            ]
        )

    @classmethod
    def from_author(cls, a, wait=True):
        """
        Parameters
        ----------
        a: :class:`~gismap.sources.hal.HALAuthor`
            Hal researcher.
        wait: :class:`bool`, default=True
            Wait a bit to avoid 429.

        Returns
        -------
        :class:`list`
            Papers available in HAL.

        Examples
        --------

        >>> fabien = HAL.search_author("Fabien Mathieu")[0]
        >>> publications = sorted(fabien.get_publications(), key=lambda p: p.title)
        >>> publications[2] # doctest:  +NORMALIZE_WHITESPACE
        HALPublication(title='Achievable Catalog Size in Peer-to-Peer Video-on-Demand Systems',
        authors=[HALAuthor(name='Yacine Boufkhad', key='yacine-boufkhad'),
        HALAuthor(name='Fabien Mathieu', key='fabien-mathieu'),
        HALAuthor(name='Fabien de Montgolfier', key='949013', key_type='pid'),
        HALAuthor(name='Diego Perino', key='Diego Perino', key_type='fullname'),
        HALAuthor(name='Laurent Viennot', key='laurentviennot')],
        venue='Proceedings of the 7th Internnational Workshop on Peer-to-Peer Systems (IPTPS)', type='conference',
        year=2008, key='471724')
        >>> diego = publications[2].authors[3]
        >>> diego
        HALAuthor(name='Diego Perino', key='Diego Perino', key_type='fullname')
        >>> len(diego.get_publications()) > 28
        True
        >>> publications[-7] # doctest:  +NORMALIZE_WHITESPACE
        HALPublication(title='Upper bounds for stabilization in acyclic preference-based systems',
        authors=[HALAuthor(name='Fabien Mathieu', key='fabien-mathieu')],
        venue="SSS'07 - 9th international conference on Stabilization, Safety, and Security of Distributed Systems",
        type='conference', year=2007, key='668356')

        Case of someone with multiple ids one want to cumulate:

        >>> maria = HAL.search_author('Maria Potop-Butucaru')
        >>> maria  # doctest: +NORMALIZE_WHITESPACE
        [HALAuthor(name='Maria Potop-Butucaru', key='841868', key_type='pid')]
        >>> n_pubs = len(HAL.from_author(maria[0]))
        >>> n_pubs > 200
        True
        >>> n_pubs == len(maria[0].get_publications())
        True

        Note: an error is raised if not enough data is provided

        >>> HAL.from_author(HALAuthor('Fabien Mathieu'))
        Traceback (most recent call last):
        ...
        ValueError: HALAuthor(name='Fabien Mathieu') must have a key for publications to be fetched.
        """
        api = "https://api.archives-ouvertes.fr/search/"
        fields = [
            "docid",
            "abstract_s",
            "label_s",
            "uri_s",
            "*Title_s",
            "title_s",
            "producedDateY_i",
            "auth_s",
            "authFullNamePersonIDIDHal_fs",
            "docType_s",
        ]
        params = {"fl": fields, "rows": 2000, "wt": "json"}
        if a.key is None:
            raise ValueError(f"{a} must have a key for publications to be fetched.")
        if a.key_type == "pid":
            params["q"] = f"authIdPerson_i:{a.key}"
        elif a.key_type == "fullname":
            params["q"] = f'authFullName_s:"{a.key}"'
        else:
            params["q"] = f"authIdHal_s:{a.key}"
        r = get(api, params=params)
        response = json.loads(r)["response"]
        res = [HALPublication.from_json(r) for r in response.get("docs", [])]
        if wait:
            sleep(cls.author_backoff)
        if len(res) == 0 and a.key_type != "fullname":
            name = a.name
            return HAL.from_author(HALAuthor(name=name, key=name, key_type="fullname"))
        return res


@dataclass(repr=False)
class HALAuthor(Author, HAL):
    """
    Author from the HAL (Hyper Articles en Ligne) database.

    HAL is a French open archive for scholarly publications.

    Parameters
    ----------
    name: :class:`str`
        The author's name.
    key: :class:`str` or :class:`int`, optional
        HAL identifier for the author.
    key_type: :class:`str`, optional
        Type of key ('pid', 'fullname', or None for idHal).
    aliases: :class:`list`
        Alternative names for the author.
    """

    key: str | int = None
    key_type: str = None
    aliases: list = field(default_factory=list)
    _url: str = None
    _img: str = None
    _cv: bool = None

    def __post_init__(self):
        if self.key and self.key_type is None:
            if self.key.isdigit():
                self.key_type = "pid"
            if " " in self.key:
                self.key_type = "fullname"

    def check_cv(self):
        if self.key_type is not None:
            self._cv = False
            return None
        url = f"https://cv.hal.science/{self.key}"
        soup = Soup(get(url), "lxml")
        if soup.form:
            self._cv = False
            return None
        self._cv = True
        self._url = url
        try:
            self._img = soup.main.section.div.div.div.img["src"]
        except TypeError:
            return None

    @property
    def url(self):
        if self._url is not None:
            return self._url
        if self.key_type == "pid":
            return f"https://hal.science/search/index/?q=*&authIdPerson_i={self.key}"
        elif self.key_type == "fullname":
            return f"https://hal.science/search/index?q={quote_plus(self.name)}"
        else:
            return f"https://hal.science/search/index/?q=*&authIdHal_s={self.key}"

    @property
    def img(self):
        if self._cv is None:
            self.check_cv()
        return self._img

    def get_publications(self):
        return HAL.from_author(self)


def parse_facet_author(a):
    """

    Parameters
    ----------
    a: :class:`str`
        Hal facet of author

    Returns
    -------
    :class:`~gismap.sources.hal.HALAuthor`

    """
    name, pid, hid = a.split("_FacetSep_")
    if hid:
        return HALAuthor(name=name, key=hid)
    elif pid and int(pid):
        return HALAuthor(name=name, key=pid, key_type="pid")
    else:
        return HALAuthor(name=name, key=name, key_type="fullname")


HAL_TYPES = {
    "ART": "journal",
    "COMM": "conference",
    "OUV": "book",
    "COUV": "chapter",
    "THESE": "thesis",
    "UNDEFINED": "report",
}

HAL_KEYS = {
    "title_s": "title",
    "abstract_s": "abstract",
    "docid": "key",
    "bookTitle_s": "booktitle",
    "conferenceTitle_s": "conference",
    "journalTitle_s": "journal",
    "docType_s": "type",
    "producedDateY_i": "year",
    "uri_s": "url",
}


@dataclass(repr=False)
class HALPublication(Publication, HAL):
    """
    Publication from the HAL database.

    Parameters
    ----------
    title: :class:`str`
        Publication title.
    authors: :class:`list`
        List of :class:`HALAuthor` objects.
    venue: :class:`str`
        Publication venue.
    type: :class:`str`
        Publication type.
    year: :class:`int`
        Publication year.
    key: :class:`str`
        HAL document identifier.
    metadata: :class:`dict`
        Additional metadata (abstract, URL, etc.).
    """

    key: str
    metadata: dict = field(default_factory=dict)

    @property
    def url(self):
        return self.metadata.get("url")

    @classmethod
    def from_json(cls, r):
        """

        Parameters
        ----------
        r: :class:`dict`
            De-serialized JSON.

        Returns
        -------
        :class:`~gismap.sources.hal.HALPublication`

        """
        keys = {v: unlist(r[k]) for k, v in HAL_KEYS.items() if k in r}
        res = {k: keys[k] for k in ["key", "title", "year"]}
        # res = {v: unlist(r[k]) for k, v in HAL_KEYS.items() if k in r}
        res["authors"] = [parse_facet_author(a) for a in r.get("authFullNamePersonIDIDHal_fs", [])]
        for tag in ["booktitle", "journal", "conference"]:
            if tag in keys:
                res["venue"] = keys[tag]
                break
        else:
            res["venue"] = "unpublished"
        res["type"] = HAL_TYPES.get(keys["type"], keys["type"].lower())
        res["metadata"] = {k: keys[k] for k in {"abstract", "url"} if k in keys and keys[k]}
        return cls(**res)



import re
from dataclasses import dataclass, field

from gismap.sources.models import DB, db_class_to_auth_class
from gismap.sources.multi import SourcedAuthor, sort_author_sources
from gismap.utils.common import LazyRepr, get_classes, list_of_objects
from gismap.utils.logger import logger

default_dbs = ["hal", "ldb"]


def db_dict():
    """Lazy lookup of DB subclasses (avoids import-order dependency)."""
    return get_classes(DB, key="db_name")


@dataclass(repr=False)
class AuthorMetadata(LazyRepr):
    """
    Optional information about an author to be used to enhance her presentation.

    Attributes
    ----------

    url: :class:`str`
        Homepage of the author.
    img: :class:`str`
        Url to a picture.
    group: :class:`str`
        Group of the author.
    position: :class:`tuple`
        Coordinates of the author.
    """

    url: str = None
    img: str = None
    group: str = None
    position: tuple = None


@dataclass(repr=False)
class LabAuthor(SourcedAuthor):
    """
    Examples
    --------
    The metadata and DB key(s) of an author can be entered in parentheses using key/values.

    Improper key/values are ignored (with a warning).

    >>> dummy= LabAuthor("My Name(img: https://my.url.img, group:me,url:https://mysite.org,hal:key1,ldb:toto,badkey:hello,no_colon_separator)")
    >>> dummy.metadata
    AuthorMetadata(url='https://mysite.org', img='https://my.url.img', group='me')
    >>> dummy.sources
    [HALAuthor(name='My Name', key='key1'), LDBAuthor(name='My Name', key='toto')]

    You can enter multiple keys for the same DB. HAL key types are automatically detected.

    >>> dummy2= LabAuthor("My Name (hal:key1,hal:123456,hal: My Other Name )")
    >>> dummy2.sources  # doctest: +NORMALIZE_WHITESPACE
    [HALAuthor(name='My Name', key='key1'),
    HALAuthor(name='My Name', key='123456', key_type='pid'),
    HALAuthor(name='My Name', key='My Other Name', key_type='fullname')]
    """

    metadata: AuthorMetadata = field(default_factory=AuthorMetadata)

    @property
    def fingerprint(self):
        return self.key if self.key is not None else super().fingerprint

    def auto_img(self):
        for source in self.sources:
            img = getattr(source, "img", None)
            if img is not None:
                self.metadata.img = img
                break

    def __post_init__(self):
        pattern = r"\s*([^,(]+)\s*(?:\(([^)]*)\))?\s*$"
        match = re.match(pattern, self.name)
        if match:
            self.name = match.group(1).strip()
            content = match.group(2)
            if content:
                for kv in content.split(","):
                    if ":" not in kv:
                        logger.warning(f"I don't know what to do with {kv}.")
                        continue
                    k, v = kv.split(":", 1)
                    k = k.strip().lower()
                    v = v.strip()
                    if k in db_dict():
                        DBAuthor = db_class_to_auth_class(db_dict()[k])
                        self.sources.append(DBAuthor(name=self.name, key=v))
                    elif k in ["url", "img", "group"]:
                        setattr(self.metadata, k, v)
                    else:
                        logger.warning(f"I don't know what to do with {kv}.")
        else:
            self.name = self.name.strip()

    def auto_sources(self, dbs=None):
        """
        Automatically populate the sources based on author's name.

        Parameters
        ----------
        dbs: :class:`list`, default=[:class:`~gismap.sources.hal.HAL`, :class:`~gismap.sources.dblp.DBLP`]
            List of DB sources to use.

        Returns
        -------
        None
        """
        dbs = list_of_objects(dbs, db_dict(), default=default_dbs)
        sources = []
        for db in dbs:
            source = db.search_author(self.name)
            if len(source) == 0:
                logger.info(f"{self.name} not found in {db.db_name}")
            elif len(source) > 1:
                logger.info(f"Multiple entries for {self.name} in {db.db_name}")
            sources += source
        if len(sources) > 0:
            self.sources = sort_author_sources(sources)


def labify_author(author, rosetta):
    """
    Convert a database author to a LabAuthor if possible.

    Parameters
    ----------
    author : :class:`~gismap.sources.models.Author`
        Author to convert.
    rosetta : :class:`dict`
        Mapping from keys/names to LabAuthor objects.

    Returns
    -------
    :class:`~gismap.lab.lab_author.LabAuthor` or original author
        LabAuthor if found in rosetta, otherwise the original author.
    """
    if isinstance(author, LabAuthor):
        return author
    return rosetta.get(author.key, rosetta.get(author.name, author))


def labify_publications(pubs, rosetta):
    """
    Convert publication authors to LabAuthors in place.

    Parameters
    ----------
    pubs : :class:`list`
        Publications to update.
    rosetta : :class:`dict`
        Mapping from keys/names to LabAuthor objects.

    Returns
    -------
    None
    """
    for pub in pubs:
        pub.authors = [labify_author(a, rosetta) for a in pub.authors]
        for source in getattr(pub, "sources", []):
            source.authors = [labify_author(a, rosetta) for a in pub.authors]



from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.gismo import Gismo
from sklearn.feature_extraction.text import CountVectorizer


def make_post_publi(lab):
    """
    Hook to turn publication key stored in a corpus into actual publication.

    Parameters
    ----------
    lab: :class:`~gismap.lab.labmap.LabMap`
        Lab that contains the corpus publications.

    Returns
    -------
    callable
    """

    def to_bib(g, i):
        item = g.corpus[i]
        return lab.publications[item]

    return to_bib


def make_gismo(lab, vectorizer_parameters=None):
    """
    Makes a gismo out of a lab.

    Parameters
    ----------
    lab: :class:`~gismap.lab.labmap.LabMap`
        Lab that contains publications.
    vectorizer_parameters: :class:`dict`
        Overriding parameters for the Countvectorizer of the gismo.

    Returns
    -------
    gismo: :class:`~gismo.gismo.Gismo`
        Gismo of the lab.
    """
    parameters = {"ngram_range": (1, 3), "dtype": float, "stop_words": sw, "min_df": 3}
    if vectorizer_parameters is not None:
        parameters.update(vectorizer_parameters)
    corpus = Corpus([p for p in lab.publications], to_text=lab.publi_to_text)
    vectorizer = CountVectorizer(**parameters)
    embedding = Embedding(vectorizer=vectorizer)
    embedding.fit_transform(corpus)
    gismo = Gismo(corpus, embedding)
    gismo.post_documents_item = make_post_publi(lab)
    return gismo


stop_words = [
    "01",
    "20plus",
    "academia",
    "academic",
    "academy",
    "académie",
    "acm",
    "activities",
    "actualités",
    "adresse",
    "advances",
    "affichertoutesdepuis",
    "ajouter",
    "an",
    "ancitations",
    "and",
    "annual",
    "annéetrier",
    "antipolis",
    "are",
    "article",
    "articledisponiblesnon",
    "articles",
    "articles0",
    "arxiv",
    "as",
    "astérisque",
    "at",
    "attended",
    "au",
    "auteuradresse",
    "auteurnouveaux",
    "auteurnouvelles",
    "aux",
    "award",
    "awarded",
    "awards",
    "base",
    "based",
    "be",
    "been",
    "bibliography",
    "bibliothèquemétriquesalertesparamètresconnexionconnexionobtenir",
    "board",
    "book",
    "born",
    "by",
    "california",
    "called",
    "can",
    "celles",
    "centrale",
    "centre",
    "ces",
    "cet",
    "cette",
    "chair",
    "chaussées",
    "ciarlet",
    "citations",
    "citationstrier",
    "cited",
    "citée",
    "cnrs",
    "cnrsadresse",
    "coauteurscoauteurssuivre",
    "collaboration",
    "colloquium",
    "collège",
    "columbia",
    "committee",
    "comptabilisées",
    "conference",
    "contact",
    "contributions",
    "council",
    "cours",
    "course",
    "courses",
    "dans",
    "de",
    "des",
    "diegoadresse",
    "différentes",
    "director",
    "disponiblessur",
    "doi",
    "données",
    "doubleles",
    "du",
    "décompte",
    "early",
    "earned",
    "ecole",
    "ed",
    "edinburgh",
    "elected",
    "en",
    "english",
    "ens",
    "envoi",
    "et",
    "europaea",
    "european",
    "events",
    "exigences",
    "fellow",
    "financementcoauteurstout",
    "for",
    "formerly",
    "forum",
    "fr",
    "france",
    "franceadresse",
    "français",
    "françois",
    "french",
    "from",
    "fusionnéesle",
    "fusionnés",
    "google",
    "grenoble",
    "gérard",
    "habilitation",
    "had",
    "he",
    "her",
    "here",
    "highly",
    "his",
    "home",
    "hong",
    "honorary",
    "ici",
    "ieee",
    "imag",
    "in",
    "inclut",
    "informatique",
    "innovation",
    "inria",
    "insa",
    "institut",
    "international",
    "invited",
    "is",
    "isbn",
    "it",
    "je",
    "jean",
    "jour",
    "journal",
    "july",
    "known",
    "kong",
    "la",
    "lab",
    "le",
    "lecture",
    "les",
    "lille",
    "liées",
    "liés",
    "lncs",
    "lyon",
    "mail",
    "maintenant",
    "medal",
    "media",
    "member",
    "mises",
    "mon",
    "monde",
    "my",
    "mécanique",
    "national",
    "ne",
    "nombre",
    "normale",
    "notes",
    "notificationsokmon",
    "novel",
    "of",
    "olivier",
    "on",
    "opération",
    "ordre",
    "page",
    "pages",
    "paper",
    "par",
    "paraccès",
    "parcitée",
    "paris",
    "paristech",
    "partout",
    "pas",
    "pdf",
    "peut",
    "peuvent",
    "peux",
    "ph",
    "phd",
    "pierre",
    "plus",
    "polytechnique",
    "ponts",
    "pour",
    "pp",
    "premier",
    "preprint",
    "president",
    "prix",
    "prize",
    "proceedings",
    "professor",
    "profil",
    "profilcitée",
    "profilma",
    "programme",
    "programmes",
    "propre",
    "présentation",
    "publications",
    "publiccoauteurstitretriertrier",
    "published",
    "que",
    "qui",
    "received",
    "recherche",
    "record",
    "report",
    "research",
    "researcher",
    "réaliser",
    "réessayer",
    "résultatsaideconfidentialitéconditions",
    "saclayadresse",
    "san",
    "scholar",
    "scholarchargement",
    "school",
    "sciences",
    "scientifique",
    "scientist",
    "selected",
    "senior",
    "she",
    "sigmod",
    "silver",
    "site",
    "slides",
    "sont",
    "sophia",
    "sorbonne",
    "southern",
    "speaker",
    "springer",
    "stanford",
    "sud",
    "suivants",
    "suivies",
    "summer",
    "supervision",
    "supérieure",
    "supérieureadresse",
    "sur",
    "symposium",
    "système",
    "tard",
    "temps",
    "texas",
    "that",
    "the",
    "then",
    "theses",
    "titrecitée",
    "to",
    "transactions",
    "travaux",
    "télécom",
    "un",
    "under",
    "une",
    "univ",
    "university",
    "universityadresse",
    "université",
    "upmc",
    "usa",
    "using",
    "validée",
    "verimag",
    "verlag",
    "veuillez",
    "via",
    "vol",
    "was",
    "with",
    "won",
    "worked",
    "www",
    "year",
    "école",
    "être",
    "towards",
    "paradigm",
    "we",
    "this",
    "which",
    "our",
    "proposed",
    "their",
    "approach",
    "each",
    "such",
    "show",
    "what",
    "nous",
]

sw = stop_words + [str(i) for i in range(2100)]



"""
Build script for creating the LDB (Local DBLP) database.

Run as a module to download and process the DBLP dataset:

    python -m gismap.build

This will fetch the DBLP RDF dump from the website and create a compressed local database.
"""

if __name__ == "__main__":
    import argparse

    from gismap.sources.ldb import LDB

    parser = argparse.ArgumentParser(description="Build LDB database from DBLP TTL dump.")
    parser.add_argument(
        "--no-search",
        action="store_true",
        help="Exclude search engine from output (for GitHub assets).",
    )
    args = parser.parse_args()

    LDB.build_db()
    LDB.dump_db(include_search=not args.no_search)



from collections import defaultdict
from itertools import combinations

import numpy as np
from domonic import tags


def initials(name):
    """
    Parameters
    ----------
    name: :class:`str`
        Person's name.

    Returns
    -------
    :class:`str`
        Person's initials (2 letters only).
    """
    first_letters = [w[0] for w in name.split()]
    return first_letters[0].upper() + first_letters[-1].upper()


def linkify(name, url):
    """
    Wrap a name in an HTML link if URL is provided.

    Parameters
    ----------
    name : :class:`str`
        Display text.
    url : :class:`str` or None
        Target URL, or None for no link.

    Returns
    -------
    :class:`str`
        HTML anchor tag or span.
    """
    if url:
        return f'<a href="{url}" target="_blank">{name}</a>'
    else:
        return f"<span>{name}</span>"


def author_html(author):
    """
    Parameters
    ----------
    author: :class:`~gismap.sources.models.Author`
        Author.

    Returns
    -------
    HTML string with URL if applicable.
    """
    name = getattr(author, "name", "Unknown Author")
    # Try direct URL property (optional)
    url = getattr(author, "url", None)
    # For LabAuthor, check metadata.url
    if hasattr(author, "metadata"):
        meta_url = getattr(author.metadata, "url", None)
        if meta_url:
            url = meta_url
        elif hasattr(author.sources[0], "url"):
            url = author.sources[0].url
    return linkify(name, url)


def pub_html(pub):
    """
    Parameters
    ----------
    pub: :class:`~gismap.sources.models.Publication`
        Publication.

    Returns
    -------
    HTML string with hyperlinks where applicable.
    """
    # Title as link if available
    url = getattr(pub, "url", None)
    if url is None and hasattr(pub, "sources"):
        url = getattr(pub.sources[0], "url", None)
    title_html = linkify(pub.title, url)

    # Authors: render in order, separated by comma
    authors_html = ", ".join([author_html(author) for author in getattr(pub, "authors", [])])

    # Venue, Year
    venue = getattr(pub, "venue", "")
    year = getattr(pub, "year", "")

    # Basic HTML layout
    html = f"{title_html}, by <i>{authors_html}</i>. {venue}, {year}."
    return html.strip()


expand_script = """var elts = this.parentElement.parentElement.querySelectorAll('.extra-publication');
for (var i = 0; i < elts.length; ++i) {elts[i].style.display = 'list-item';}
this.parentElement.style.display = 'none';
return false;"""


def publications_list(publications, n=10):
    """
    Parameters
    ----------
    publications: :class:`list` of :class:`~gismap.sources.models.Publication`
        Publications to display.
    n: :class:`int`, default=10
        Number of publications to display. If there are more publications,
        a *Show more* option is available to unravel them.

    Returns
    -------
    :class:`~domonic.html.ul`
    """
    lis = []
    for i, pub in enumerate(publications):
        if i < n:
            lis.append(f"<li>{pub_html(pub)}</li>")
        else:
            lis.append(f'<li class="extra-publication" style="display:none;">{pub_html(pub)}</li>')
    if len(publications) > n:
        lis.append(f'<li><a href="#" onclick="{expand_script}">Show more…</a></li>')
    return "<ul>\n" + "\n".join(lis) + "</ul>\n"


def to_node(s, node_pubs):
    """
    Parameters
    ----------
    s: :class:`~gismap.lab.lab_author.LabAuthor`
        Author.
    node_pubs: :class:`dict`
        Lab publications.

    Returns
    -------
    :class:`dict`
        A display-ready representation of the author.
    """
    overlay = tags.div()
    overlay.appendChild(tags.div(f"Publications of {author_html(s)}"))
    overlay.appendChild(tags.div(publications_list(node_pubs[s.key])))

    res = {
        "id": s.key,
        "hover": f"Click for details on {s.name}.",
        "overlay": f"{overlay}",
        "group": s.metadata.group,
    }
    if s.metadata.img:
        res.update({"image": s.metadata.img, "shape": "circularImage"})
    else:
        res["label"] = initials(s.name)
    if s.metadata.position:
        x, y = s.metadata.position
        res.update({"x": x, "y": y, "fixed": True})
    return res


def to_edge(k, v, searchers):
    """
    Parameters
    ----------
    k: :class:`tuple`
        Keys of the authors involved.
    v: :class:`list`
        List of joint publications.
    authors: :class:`dict`
        Authors.

    Returns
    -------
    :class:`dict`
        A display-ready representation of the collaboration edge.
    """
    strength = 1 + np.log2(len(v))
    overlay = tags.div()
    overlay.appendChild(
        tags.div(f"Joint publications from {author_html(searchers[k[0]])} and {author_html(searchers[k[1]])}:")
    )
    overlay.appendChild(tags.div(f"{publications_list(v)}"))
    res = {
        "from": k[0],
        "to": k[1],
        "hover": f"Show joint publications from {searchers[k[0]].name} and {searchers[k[1]].name}",
        "overlay": f"{overlay}",
        "width": int(strength),
        "length": int(200 / strength),
    }
    g1, g2 = searchers[k[0]].metadata.group, searchers[k[1]].metadata.group
    if g1 and g2 and g1 != g2:
        res["color"] = "rgba(0,0,0,0)"
    return res


def lab_to_graph(lab):
    """
    Parameters
    ----------
    lab: :class:`~gismap.lab.labmap.LabMap`
        A lab populated with authors and publications.

    Returns
    -------
    :class:`str`
        Collaboration graph.

    Examples
    --------

    >>> from gismap.lab import ListMap as Map
    >>> lab = Map(author_list=['Tixeuil Sébastien', 'Mathieu Fabien'], name='mini', dbs="hal")
    >>> lab.update_authors()
    >>> lab.update_publis()
    >>> len(lab.authors)
    2
    >>> 327 < len(lab.publications) < 430
    True
    >>> nodes, edges = lab_to_graph(lab)
    >>> nodes[0]['group']
    'mini'
    >>> edges[0]['hover']
    'Show joint publications from Mathieu Fabien and Tixeuil Sébastien'
    >>> html = lab.html(groups={"mini": {"color": "#777"}})
    """
    node_pubs = defaultdict(list)  # {k: [] for k in lab.authors}
    edges_dict = defaultdict(list)
    for p in lab.publications.values():
        # Strange things can happen with multiple sources. This should take care of it.
        lauths = {a.key: a for source in p.sources for a in source.authors if a.__class__.__name__ == "LabAuthor"}
        lauths = sorted([a for a in lauths.values()], key=lambda a: str(a.key))
        for a in lauths:
            node_pubs[a.key].append(p)
        for a1, a2 in combinations(lauths, 2):
            edges_dict[a1.key, a2.key].append(p)
    # connected = {k for kl in edges_dict for k in kl}
    for k, v in node_pubs.items():
        node_pubs[k] = sorted(v, key=lambda p: -p.year)
    for k, v in edges_dict.items():
        edges_dict[k] = sorted(v, key=lambda p: -p.year)
    nodes = [
        to_node(s, node_pubs)
        for s in lab.authors.values()  # if s.key in connected
    ]
    edges = [to_edge(k, v, lab.authors) for k, v in edges_dict.items()]
    # for node in nodes:
    #     node['connected'] = node['id'] in connected

    return nodes, edges



import errno
import json
import os
from dataclasses import dataclass, field
from datetime import UTC, datetime
from functools import lru_cache
from importlib.metadata import version as pkg_version
from pathlib import Path
from typing import ClassVar

import dill as pickle
import numba as nb
import numpy as np
import requests
import zstandard as zstd
from bof.fuzz import Process
from gismo.common import safe_write
from platformdirs import user_data_dir
from tqdm.auto import tqdm

from gismap.sources.dblp_ttl import publis_streamer
from gismap.sources.models import DB, Author, Publication
from gismap.utils.common import Data
from gismap.utils.logger import logger
from gismap.utils.text import normalized_name
from gismap.utils.zlist import ZList

DATA_DIR = Path(
    user_data_dir(
        appname="gismap",
        appauthor=False,
    )
)
LDB_STEM = "ldb"
GITHUB_REPO = "balouf/gismap"

LDB_PARAMETERS = Data(
    {
        "search": {"limit": 3, "cutoff": 87.0, "slack": 1.0},
        "bof": {"n_range": 2, "length_impact": 0.1},
        "frame_size": {"authors": 512, "publis": 256},
        "io": {
            "source": "https://dblp.org/rdf/dblp.ttl.gz",
            "destination": DATA_DIR / f"{LDB_STEM}.pkl.zst",
            "metadata": DATA_DIR / f"{LDB_STEM}.json",
            "gh_api": f"https://api.github.com/repos/{GITHUB_REPO}/releases",
        },
    }
)
"""
Global configuration parameters for the Local DBLP (LDB) pipeline.

Structure:

- **search**:

  - *limit*: maximum number of candidates retrieved per query.
  - *cutoff*: minimal similarity score required to keep a candidate.
  - *slack*: tolerance around the cutoff for borderline matches.

- **bof** (Bag-of-Factors):

  - *n_range*: max factor size (higher is better but more expensive).
  - *length_impact*: how to compare two inputs of different size.

- **frame_size**:

  - *authors*: maximum number of authors kept in a single frame/batch.
  - *publis*: maximum number of publications kept in a single frame/batch.

- **io**:

  - *source*: URL/file location of the DBLP RDF dump used as raw input.
  - *destination*: local path where the compressed preprocessed dataset is / will be stored.
  - *gh_api*: GitHub API endpoint used to fetch release information for the project.

LDB_PARAMETERS is a Data (RecursiveDict) instance, so nested fields can be
accessed with attribute notation, e.g.::

    LDB_PARAMETERS.search.limit
    LDB_PARAMETERS.io.destination
"""


def _compatible_tag(tag: str) -> bool:
    """Check that release tag is compatible with the installed package (same major.minor)."""
    v = pkg_version("gismap")
    pkg_minor = ".".join(v.split(".")[:2])
    tag_clean = tag.lstrip("v")
    tag_minor = ".".join(tag_clean.split(".")[:2])
    return pkg_minor == tag_minor


@dataclass(repr=False)
class LDB(DB):
    """
    Browse DBLP from a local copy of the database.

    LDB is a class-only database - it should not be instantiated.
    All methods are classmethods and state is stored in class variables.

    Examples
    --------

    Public DB methods ensure that the DB is loaded but if you need to use a specific LDB method, prepare the DB first.

    >>> LDB._ensure_loaded()
    >>> LDB.author_by_key("66/2077")
    LDBAuthor(name='Fabien Mathieu', key='66/2077')
    >>> pubs = sorted(LDB.author_publications('66/2077'), key = lambda p: p.year)
    >>> pub = pubs[0]
    >>> pub.metadata
    {'url': 'http://www2003.org/cdrom/papers/poster/p102/p102-mathieu.htm', 'streams': ['conf/www']}
    >>> christelle = LDB.search_author("Christelle Caillouet")
    >>> christelle
    [LDBAuthor(name='Christelle Caillouet', key='10/8725')]
    >>> christelle[0].aliases
    ['Christelle Molle']
    >>> LDB.db_info()  # doctest: +SKIP
    {'tag': 'v0...', 'downloaded_at': '...', 'size': ..., 'path': ...}
    >>> LDB.check_update()
    >>> ldb = LDB()
    Traceback (most recent call last):
    ...
    TypeError: LDB should not be instantiated. Use class methods directly, e.g., LDB.search_author(name)
    """

    db_name: ClassVar[str] = LDB_STEM
    parameters: ClassVar[Data] = LDB_PARAMETERS

    # Class-level state (replaces instance attributes)
    authors: ClassVar[ZList | None] = None
    publis: ClassVar[ZList | None] = None
    keys: ClassVar[dict | None] = None
    search_engine: ClassVar[Process | None] = None
    _initialized: ClassVar[bool] = False

    __hash__ = object.__hash__

    def __init__(self):
        raise TypeError("LDB should not be instantiated. Use class methods directly, e.g., LDB.search_author(name)")

    @classmethod
    def _ensure_loaded(cls):
        """Lazy-load the database if not already loaded."""
        if cls._initialized:
            return
        if not cls.parameters.io.destination.exists():
            logger.info("LDB not found locally. Attempting to retrieve from GitHub...")
            try:
                cls.retrieve()
            except RuntimeError as e:
                logger.warning(f"Could not auto-retrieve LDB: {e}")
        cls.load_db()

    @classmethod
    def build_db(cls, limit=None):
        """
        Build the LDB database from a DBLP TTL dump.

        Parses the DBLP RDF/TTL file to extract publications and authors,
        stores them in compressed ZList structures, and builds a fuzzy
        search engine for author name lookups.

        Parameters
        ----------
        limit: :class:`int`, optional
            Maximum number of publications to process. If None, processes
            the entire database. Useful for testing with a subset.

        Notes
        -----
        This method populates the class-level attributes:

        - ``authors``: ZList of (key, name, publication_indices) tuples
        - ``publis``: ZList of publication records
        - ``keys``: dict mapping author keys to indices
        - ``search_engine``: fuzzy search Process for author lookups

        After building, call :meth:`dump_db` to persist the database.

        Examples
        --------
        Build from the default DBLP source:

        >>> LDB.build_db()  # doctest: +SKIP
        >>> LDB.dump_db()   # doctest: +SKIP

        Build a small test database:

        >>> LDB.build_db(limit=1000)
        >>> LDB.authors[0]
        ('78/459-1', ['Manish Singh'], [0])

        Save your build in a non-default file:

        >>> from tempfile import TemporaryDirectory
        >>> from pathlib import Path
        >>> with TemporaryDirectory() as tmpdirname:
        ...     LDB.dump(filename="test.zst", path=tmpdirname)
        ...     [file.name for file in Path(tmpdirname).glob("*")]
        ['test.zst']

        In case you don't like your build and want to reload your local database from disk:

        >>> LDB.load_db()
        """
        source = cls.parameters.io.source
        authors_dict = dict()
        logger.info("Retrieve publications")
        with ZList(frame_size=cls.parameters.frame_size.publis) as publis:
            for i, (
                key,
                title,
                typ,
                authors,
                url,
                streams,
                pages,
                venue,
                year,
            ) in enumerate(publis_streamer(source)):
                auth_indices = []
                for auth_key, auth_name in authors.items():
                    if auth_key not in authors_dict:
                        authors_dict[auth_key] = (len(authors_dict), {auth_name}, [i])
                    else:
                        authors_dict[auth_key][1].add(auth_name)
                        authors_dict[auth_key][2].append(i)
                    auth_indices.append(authors_dict[auth_key][0])
                publis.append((key, title, typ, auth_indices, url, streams, pages, venue, year))
                if i == limit:
                    break
        cls.publis = publis
        logger.info(f"{len(publis)} publications retrieved.")
        logger.info("Compact authors")
        with ZList(frame_size=cls.parameters.frame_size.authors) as authors:
            for key, (_, names, pubs) in tqdm(authors_dict.items()):
                authors.append((key, list(names), pubs))
        cls.authors = authors
        cls.keys = {k: v[0] for k, v in authors_dict.items()}
        del authors_dict
        cls._build_search_engine()
        cls._invalidate_cache()
        cls._initialized = True

    @classmethod
    def _build_search_engine(cls):
        cls.search_engine = Process(
            n_range=cls.parameters.bof.n_range,
            length_impact=cls.parameters.bof.length_impact,
        )
        main_names = []
        aliases = []
        aliases_indices = []
        for i, a in enumerate(cls.authors):
            main_names.append(normalized_name(a[1][0]))
            for alias in a[1][1:]:
                aliases_indices.append(i)
                aliases.append(normalized_name(alias))
        cls.search_engine.fit(main_names + aliases)
        aliases_indices = np.array(aliases_indices)
        cls.search_engine.choices = np.concatenate((np.arange(len(cls.authors)), aliases_indices))
        # cls.search_engine.fit([normalized_name(a[1]) for a in cls.authors])
        # cls.search_engine.choices = np.arange(len(cls.authors))
        cls.search_engine.vectorizer.features_ = cls.numbify_dict(cls.search_engine.vectorizer.features_)
        logger.info(f"{len(cls.authors)} authors indexed.")

    @classmethod
    @lru_cache(maxsize=50000)
    def author_by_index(cls, i):
        key, names, _ = cls.authors[i]
        names = sorted(names)
        return LDBAuthor(key=key, name=names[0], aliases=names[1:])

    @classmethod
    def author_by_key(cls, key):
        return cls.author_by_index(cls.keys[key])

    @classmethod
    @lru_cache(maxsize=50000)
    def publication_by_index(cls, i):
        key, title, typ, authors, url, streams, pages, venue, year = cls.publis[i]
        if venue is None:
            venue = "unpublished"
        return {
            "key": key,
            "title": title,
            "type": typ,
            "authors": authors,
            "url": url,
            "streams": streams,
            "pages": pages,
            "venue": venue,
            "year": year,
        }

    @classmethod
    def author_publications(cls, key):
        cls._ensure_loaded()
        _, name, pubs = cls.authors[cls.keys[key]]
        pubs = [cls.publication_by_index(k).copy() for k in pubs]
        auth_ids = sorted({k for p in pubs for k in p["authors"]})
        auths = {k: cls.author_by_index(k) for k in auth_ids}
        for pub in pubs:
            pub["authors"] = [auths[k] for k in pub["authors"]]
            metadata = dict()
            for k in ["url", "streams", "pages"]:
                v = pub.pop(k)
                if v is not None:
                    metadata[k] = v
            pub["metadata"] = metadata
        return [LDBPublication(**pub) for pub in pubs]

    @classmethod
    @lru_cache(maxsize=1000)
    def search_author(cls, name):
        cls._ensure_loaded()
        res = cls.search_engine.extract(
            normalized_name(name),
            limit=cls.parameters.search.limit,
        )
        if not res:
            return []
        target = max(cls.parameters.search.cutoff, res[0][1] - cls.parameters.search.slack)
        res = [r[0] for r in res if r[1] > target]
        sorted_ids = {i: cls.author_by_index(i) for i in sorted(res)}
        return [sorted_ids[i] for i in res]

    @classmethod
    def _invalidate_cache(cls):
        cls.search_author.cache_clear()
        cls.publication_by_index.cache_clear()
        cls.author_by_index.cache_clear()

    @classmethod
    def from_author(cls, a):
        return cls.author_publications(a.key)

    @classmethod
    def _get_release_info(cls, tag: str | None = None) -> dict:
        """
        Fetch release metadata from GitHub API.

        Parameters
        ----------
        tag: :class:`str`, optional
            Specific release tag (e.g., "v0.4.0"). If None, fetches latest.

        Returns
        -------
        :class:`dict`
            Release metadata including tag_name and assets.

        Raises
        ------
        :class:`RuntimeError`
            If release not found or API request fails.
        """
        api_url = cls.parameters.io.gh_api
        if tag is None:
            url = f"{api_url}/latest"
        else:
            url = f"{api_url}/tags/{tag}"

        try:
            response = requests.get(url, timeout=30)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            if response.status_code == 404:
                raise RuntimeError(f"Release not found: {tag or 'latest'}") from e
            raise RuntimeError(f"GitHub API error: {e}") from e
        except requests.exceptions.RequestException as e:
            raise RuntimeError(f"Network error fetching release info: {e}") from e

    @classmethod
    def _download_file(cls, url: str, dest: Path, desc: str = "Downloading"):
        """
        Download file with progress bar.

        Parameters
        ----------
        url : str
            URL to download from.
        dest : Path
            Destination file path.
        desc : str
            Description for progress bar.
        """
        dest.parent.mkdir(parents=True, exist_ok=True)

        response = requests.get(url, stream=True, timeout=30)
        response.raise_for_status()

        total_size = int(response.headers.get("content-length", 0))

        with (
            open(dest, "wb") as f,
            tqdm(
                desc=desc,
                total=total_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as pbar,
        ):
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
                    pbar.update(len(chunk))

    @classmethod
    def _save_meta(cls, tag: str, url: str, size: int):
        """Save version metadata to JSON file."""
        meta = {
            "tag": tag,
            "url": url,
            "size": size,
            "downloaded_at": datetime.now(UTC).isoformat(),
        }
        meta_path = cls.parameters.io.metadata
        meta_path.parent.mkdir(parents=True, exist_ok=True)
        with open(meta_path, "w") as f:
            json.dump(meta, f, indent=2)

    @classmethod
    def _load_meta(cls) -> dict | None:
        """Load version metadata from JSON file."""
        meta_path = cls.parameters.io.metadata
        if not meta_path.exists():
            return None
        try:
            with open(meta_path) as f:
                return json.load(f)
        except (OSError, json.JSONDecodeError):
            return None

    @classmethod
    def retrieve(cls, version: str | None = None, force: bool = False):
        """
        Download LDB database from GitHub releases.

        Parameters
        ----------
        version: :class:`str`, optional
            Specific release version (e.g., "v0.4.0" or "0.4.0").
            If None, downloads from latest release.
        force: :class:`bool`, default=False
            Download even if same version is installed.

        Examples
        --------

        The following will get you a LDB if you do not have one.

        >>> LDB.retrieve()           # Latest compatible release  # doctest: +SKIP
        >>> LDB.retrieve("v0.5.0")   # Specific version  # doctest: +SKIP
        >>> LDB.retrieve("0.5.0")    # Also works without 'v' prefix  # doctest: +SKIP

        Of course, the tag/version must be LDB-ready.

        >>> LDB.retrieve("v0.3.0")   # Too old for LDB
        Traceback (most recent call last):
        ...
        RuntimeError: Asset 'ldb.pkl.zst' not found in release v0.3.0. Available assets: []

        Raises
        ------
        RuntimeError
            If release or asset not found, download fails, or version is incompatible.
        """
        # Normalize version string (add "v" prefix if missing)
        tag = None
        if version is not None:
            tag = version if version.startswith("v") else f"v{version}"

        # Fetch release info
        logger.info(f"Fetching release info for: {tag or 'latest'}")
        release_info = cls._get_release_info(tag)
        release_tag = release_info["tag_name"]

        # Check version compatibility (only when no explicit version requested)
        if tag is None and not _compatible_tag(release_tag):
            raise RuntimeError(
                f"Latest release {release_tag} is not compatible with installed gismap {pkg_version('gismap')}. "
                f"Use retrieve(version='v...') to force a specific version."
            )

        destination = cls.parameters.io.destination

        # Check if already installed (unless force=True)
        if not force:
            meta = cls._load_meta()
            if meta and meta.get("tag") == release_tag and destination.exists():
                logger.info(f"LDB version {release_tag} already installed. Use force=True to re-download.")
                return

        # Find ldb.pkl.zst asset in release
        assets = release_info.get("assets", [])
        ldb_asset = None
        for asset in assets:
            if asset["name"] == destination.name:
                ldb_asset = asset
                break

        if ldb_asset is None:
            raise RuntimeError(
                f"Asset '{destination.name}' not found in release {release_tag}. "
                f"Available assets: {[a['name'] for a in assets[:3]]}"
            )

        download_url = ldb_asset["browser_download_url"]
        asset_size = ldb_asset["size"]

        logger.info(f"Downloading LDB from release {release_tag} ({asset_size / 1e9:.2f} GB)")

        # Download with progress bar
        cls._download_file(download_url, destination, desc=f"LDB {release_tag}")

        # Save version metadata
        cls._save_meta(release_tag, download_url, asset_size)

        # Load database and rebuild search engine locally
        cls.load_db(restore_search=True)

        logger.info(f"LDB {release_tag} successfully installed to {destination}")

    @classmethod
    def db_info(cls) -> dict | None:
        """
        Return installed version info.

        Returns
        -------
        :class:`dict` or :class:`None`
            Dictionary with tag, date, size, path; or None if not installed.
        """
        meta = cls._load_meta()
        destination = cls.parameters.io.destination
        if meta is None or not destination.exists():
            return None

        return {
            "tag": meta.get("tag"),
            "downloaded_at": meta.get("downloaded_at"),
            "size": meta.get("size"),
            "path": str(destination),
        }

    @classmethod
    def check_update(cls) -> dict | None:
        """
        Check if a newer version is available on GitHub.

        Returns
        -------
        :class:`dict` or None
            Dictionary with update info if available, None if up to date.
        """
        try:
            release_info = cls._get_release_info()
            latest_tag = release_info["tag_name"]

            if not _compatible_tag(latest_tag):
                logger.info(f"Latest release {latest_tag} is not compatible with gismap {pkg_version('gismap')}.")
                return None

            meta = cls._load_meta()
            current_tag = meta.get("tag") if meta else None

            if current_tag == latest_tag:
                logger.info(f"LDB is up to date: {current_tag}")
                return None

            return {
                "current": current_tag,
                "latest": latest_tag,
                "message": f"Update available: {current_tag or 'not installed'} -> {latest_tag}",
            }
        except RuntimeError as e:
            logger.warning(f"Could not check for updates: {e}")
            return None

    @classmethod
    def dump(cls, filename: str, path=".", overwrite=False, include_search=True):
        """Save class state to file."""
        # Convert numba dict to regular dict for pickling
        nb_dict = None
        if include_search and cls.search_engine is not None:
            nb_dict = cls.search_engine.vectorizer.features_
            cls.search_engine.vectorizer.features_ = dict(nb_dict)

        state = {
            "authors": cls.authors,
            "publis": cls.publis,
            "keys": cls.keys,
            "search_engine": cls.search_engine if include_search else None,
        }

        # Use safe_write pattern from gismo.common
        destination = Path(path) / filename
        if destination.exists() and not overwrite:
            print(f"File {destination} already exists! Use overwrite option to overwrite.")
        else:
            with safe_write(destination) as f:
                cctx = zstd.ZstdCompressor(level=3)
                with cctx.stream_writer(f) as z:
                    pickle.dump(state, z, protocol=5)

        # Restore numba dict
        if include_search and cls.search_engine is not None:
            cls.search_engine.vectorizer.features_ = nb_dict

    @classmethod
    def load(cls, filename: str, path=".", restore_search=False):
        """Load class state from file."""
        dest = Path(path) / filename
        if not dest.exists():
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), dest)

        dctx = zstd.ZstdDecompressor()
        with open(dest, "rb") as f, dctx.stream_reader(f) as z:
            state = pickle.load(z)

        cls.authors = state["authors"]
        cls.publis = state["publis"]
        cls.keys = state["keys"]
        cls.search_engine = state["search_engine"]

        if restore_search:
            cls._build_search_engine()
            cls.dump(filename=filename, path=path, overwrite=True, include_search=True)
        elif cls.search_engine is not None:
            cls.search_engine.vectorizer.features_ = cls.numbify_dict(cls.search_engine.vectorizer.features_)

        cls._invalidate_cache()
        cls._initialized = True

    @classmethod
    def dump_db(cls, include_search=True):
        destination = cls.parameters.io.destination
        destination.parent.mkdir(parents=True, exist_ok=True)
        cls.dump(
            destination.name,
            path=destination.parent,
            overwrite=True,
            include_search=include_search,
        )

    @classmethod
    def load_db(cls, restore_search=False):
        destination = cls.parameters.io.destination
        try:
            cls.load(destination.name, path=destination.parent, restore_search=restore_search)
        except FileNotFoundError:
            logger.warning("No LDB found. Building from source...")
            cls.build_db()
            cls.dump_db()
        except TypeError as e:
            if "code expected at most" in str(e):
                logger.warning("LDB file incompatible with this Python version. Rebuilding from source...")
                cls.build_db()
                cls.dump_db()
            else:
                raise

    @classmethod
    def delete_db(cls):
        destination = cls.parameters.io.destination
        if destination.exists():
            destination.unlink()

    @staticmethod
    def numbify_dict(input_dict):
        nb_dict = nb.typed.Dict.empty(key_type=nb.types.unicode_type, value_type=nb.types.int64)
        for k, v in input_dict.items():
            nb_dict[k] = v
        return nb_dict


@dataclass(repr=False)
class LDBAuthor(Author, LDB):
    """
    Author from the LDB (Local DBLP) database.

    LDB provides local access to DBLP data without rate limiting.

    Parameters
    ----------
    name: :class:`str`
        The author's name.
    key: :class:`str`
        DBLP person identifier (pid).
    aliases: :class:`list`
        Alternative names for the author.
    """

    key: str
    aliases: list = field(default_factory=list)

    @property
    def url(self):
        return f"https://dblp.org/pid/{self.key}.html"

    def get_publications(self):
        return LDB.from_author(self)


@dataclass(repr=False)
class LDBPublication(Publication, LDB):
    """
    Publication from the LDB (Local DBLP) database.

    Parameters
    ----------
    title: :class:`str`
        Publication title.
    authors: :class:`list`
        List of :class:`LDBAuthor` objects.
    venue: :class:`str`
        Publication venue.
    type: :class:`str`
        Publication type.
    year: :class:`int`
        Publication year.
    key: :class:`str`
        DBLP record key.
    metadata: :class:`dict`
        Additional metadata (URL, streams, pages).
    """

    key: str
    metadata: dict = field(default_factory=dict)

    @property
    def url(self):
        return self.metadata.get("url", f"https://dblp.org/rec/{self.key}.html")

    @property
    def stream(self):
        if "streams" in self.metadata:
            return f"https://dblp.org/streams/{self.metadata['streams'][0]}"
        return None
